package fi.kapsi.sommite.intrans;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import android.app.Activity;
import android.content.res.AssetManager;
import android.os.Bundle;
import android.os.Handler;
import android.view.View;
import android.widget.Button;
import android.widget.TextView;

public class IntransActivity extends Activity {
	private volatile boolean active = false;
	private Button startstopbutton;
	private Handler handler;
	private Thread startstopthread=null;
	
	private void updateButtonLabel() {
		if (active) {
			startstopbutton.setText(R.string.stop);			
		} else {
			startstopbutton.setText(R.string.start);			
		}
	}
	
    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);
        handler = new Handler();
        
        active = isIntransActive();
        startstopbutton = (Button)findViewById(R.id.startstopbutton);
        updateButtonLabel();
        prepareExecutable();
        
        startstopbutton.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				if (startstopthread!=null&&startstopthread.isAlive()) return;
				if (active) {
					startstopthread=new Thread(new Runnable() { public void run() { disableIntrans(); }});
				} else {
					startstopthread=new Thread(new Runnable() { public void run() { enableIntrans(); }});
				}
				startstopthread.start();
			}
		});
    }
    public void onDestroy() {
    	super.onDestroy();
    }
    private static final int DEFAULT_BUFFER_SIZE=1024;
    private static int copy(InputStream input, OutputStream output) throws IOException{
        byte[] buffer = new byte[DEFAULT_BUFFER_SIZE];
        int count = 0;
        int n = 0;
        while (-1 != (n = input.read(buffer))) {
            output.write(buffer, 0, n);
            count += n;
        }
        return count;
    }
    
    private void errorview(String msg) {
    	final Button b = (Button)findViewById(R.id.startstopbutton);
    	final TextView t = (TextView)findViewById(R.id.errormessage);
    	t.setVisibility(View.VISIBLE);
    	b.setVisibility(View.GONE);
    	t.setText(msg);
    	final Activity a = this;
    	t.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				a.finish();
			}
		});
    }
    
    private void prepareExecutable() {
        if (active) return;
        
    	File exe = getFileStreamPath("intrans");
    	AssetManager am = getAssets();
    	InputStream si=null;
    	FileOutputStream so=null;
    		try {
    			 try {
    				 si = am.open("intrans");
    				 so = openFileOutput("intrans",MODE_PRIVATE);
        			 copy(si,so);
    			 } finally {
    	    			if (si!=null) si.close();
    	    			if (so!=null) so.close();
    			 }
    			 Process p = Runtime.getRuntime().exec(new String[]
    					 {"chmod","0755",exe.getAbsolutePath()});
    			 int n = p.waitFor();
    			 if (n!=0) { throw new Exception("Failed to make the native binary executable"); }
    		} catch (Exception e) {
    			errorview(e.getMessage());
    			return;
    		}
    }
    
    private boolean isIntransActive() {
    	File lockfile = getFileStreamPath("intrans-lock");
    	FileLock l=null;
    	RandomAccessFile fs=null;
    	try {
    		try {
    			fs = new RandomAccessFile(lockfile,"rw");
    			FileChannel c = fs.getChannel();
    			l = c.tryLock();
    			if (l==null) return true;
    			else return false;
    		} finally {
    			if (l!=null) l.release();
    			if (fs!=null) fs.close();
    		}
    	} catch (IOException e) {
    		return false;
    	}
    }
    
    private void disableIntrans() {
    	File lockfile = getFileStreamPath("intrans-lock");
    	final TextView t =(TextView)findViewById(R.id.errormessage);
		Process p=null;
    	try {
    		FileInputStream fi = new FileInputStream(lockfile);
    		Reader r = new InputStreamReader(fi,"UTF-8");
    		BufferedReader br = new BufferedReader(r);
    		String pidS=br.readLine();
    		if (pidS==null) throw new Exception("No pidfile");
    		int pid = Integer.parseInt(pidS);
    		try {
    			p = Runtime.getRuntime().exec("/system/bin/su");
    			OutputStream os=p.getOutputStream();
    			PrintStream ps=new PrintStream(os,true,"UTF-8");
    			ps.println("exec /system/bin/kill "+Integer.toString(pid));
    			ps.close();
    			int n = p.waitFor();
    			if (n!=0) throw new Exception("Killing the process failed");
    		} finally {
    			if (p!=null) p.destroy();
    		}
    	} catch (Exception e) {
    		final String msg =e.getMessage();
    		handler.post(new Runnable() { public void run() {
    			t.setText(msg);
    			t.setVisibility(View.VISIBLE);
    		}});
    		return;
    	}
    	handler.post(new Runnable() { public void run() {
    		t.setVisibility(View.GONE);
    	}});
    	
    	active=false;
    }
    
    private void executeCmd(String cmd, PrintStream p, BufferedReader i) throws Exception {
    	p.println(cmd); // must not print to stdout
    	p.println("echo $?");
    	String strResult = i.readLine();
    	
    	final String failmsg="Failed to execute: "+cmd;
    	if (strResult==null) throw new Exception(failmsg);
    	int result=Integer.parseInt(strResult);
    	if (result!=0) throw new Exception(failmsg);
    }
    
    private void installKeylayout(PrintStream p,InputStream ip) throws Exception {
    	final String layoutfile="/system/usr/keylayout/ukeyboard.kl";
    	boolean layoutExists=new File(layoutfile).exists();
    	if (!layoutExists) {
    		AssetManager am = getAssets();
    		InputStream is = am.open("ukeyboard.kl");
    		InputStreamReader ir = new InputStreamReader(is,"UTF-8");
    		BufferedReader br = new BufferedReader(ir);
    		InputStreamReader ipr = new InputStreamReader(ip,"UTF-8");
    		BufferedReader bpr = new BufferedReader(ipr);
    		executeCmd("mount -o remount,rw sys /system",p,bpr);
    		String line;
    		String redirectop="' > ";
    		while ((line=br.readLine())!=null) {
    			executeCmd("echo '"+line+redirectop+layoutfile,p,bpr);
    			redirectop="' >> ";
    		}
    		executeCmd("mount -o remount,ro sys /system",p,bpr);
    	}
    }
    
    private void enableIntrans() {
    	Process p;
    	Thread waiter=null; 
    	Thread reader=null;
    	Thread erreader=null;
    	final TextView t =(TextView)findViewById(R.id.errormessage);
    	try {
    		p = Runtime.getRuntime().exec("/system/bin/su");
    		OutputStream os=p.getOutputStream();
    		final InputStream is=p.getInputStream();
    		final InputStream es=p.getErrorStream();
    		final InputStreamReader ir = new InputStreamReader(is,"UTF-8");
    		final BufferedReader br = new BufferedReader(ir);
    		PrintStream ps=new PrintStream(os,true,"UTF-8");
    		installKeylayout(ps,is);
    		executeCmd("cd "+getFilesDir().getAbsolutePath(),ps,br);
    		ps.println("exec ./intrans");
    		ps.close();
    		final Process pp = p;
    		final BlockingQueue<String> q = new LinkedBlockingQueue<String>();
    		final String success = new String();
    		
    		waiter = new Thread(new Runnable() {
    	    	public void run() {
    	    		try {
    					int n = pp.waitFor();
    					if (n!=0) {
    						q.add("su returned a failure");
    					} else {
    						q.add("terminated unexpectedly");
    					}
    				} catch (InterruptedException e) { }
    	    	}
    		});
    		reader = new Thread(new Runnable() {
    			public void run() {
    				try {
    					String l = br.readLine();
    					if (l==null) {
    						return;
    					} else if (l.equals("Intrans started")) {
    						q.add(success);
    					} else {
    						q.add("unrecognized output string: "+l);
    					}
    				} catch (Exception e) {
    					q.add(e.getMessage());
    				}
    			}
    		});
    		erreader = new Thread(new Runnable() {
    			public void run() {
    				try {
    					byte[] buf = new byte[1024];
    					int n = es.read(buf);
    					if (n==-1) return;
    					q.add(new String(buf,0,n));
    				} catch (Exception e) {
    					q.add(e.getMessage());
    				}
    			}
    		});
    		waiter.start();
    		reader.start();
    		erreader.start();
    		
    		String i=q.take();
    		if (i!=success) throw new Exception(i);
    	} catch (Exception e) {
    		final String msg =e.getMessage();
    		handler.post(new Runnable() { public void run() {
    			t.setText(msg);
    			t.setVisibility(View.VISIBLE);
    		}});
    		return;
    	} finally {
    		if (waiter!=null) waiter.interrupt();
    		if (reader!=null) reader.interrupt();
    		if (erreader!=null) erreader.interrupt();
    	}
    	
    	handler.post(new Runnable() { public void run() {
    		t.setVisibility(View.GONE);
    	}});
    	active=true;
    }
}