#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <time.h>

#define TRYSYS(expr,msg) do { if ((expr)==-1) { perror(msg); exit(1); } else { } } while (0)
#define test_bit(bit, array)    (array[bit/8] & (1<<(bit%8)))

void init_devs(int *ifd, int *ofd)
{
  char name[80];
  char idstr[80];
  struct input_id id;
  
  TRYSYS(*ifd=open("/dev/input/event0",O_RDWR),"open event0");
  TRYSYS(ioctl(*ifd,EVIOCGNAME(sizeof(name)-1),name),"ioctl getname");  
  TRYSYS(ioctl(*ifd,EVIOCGID,&id),"ioctl getid");
  if (ioctl(*ifd,EVIOCGUNIQ(sizeof(idstr)-1),idstr)<0) {
    idstr[0]=0;
  }
  uint8_t key_bitmask[(KEY_MAX+7)/8];
  memset(key_bitmask,0,sizeof(key_bitmask));
  if (ioctl(*ifd,EVIOCGBIT(EV_KEY, sizeof(key_bitmask)), key_bitmask)<0) {
    fprintf(stderr,"/dev/input/event0 did not return a key bitmask.");
    exit(1);
  }
  name[79]=0;

  TRYSYS(*ofd=open("/dev/uinput",O_RDWR),"open uinput");
  TRYSYS(ioctl(*ofd,UI_SET_EVBIT,EV_KEY),"set evbit");
  TRYSYS(ioctl(*ofd,UI_SET_EVBIT,EV_REL),"set evbit");
  TRYSYS(ioctl(*ofd,UI_SET_RELBIT,REL_X),"set relbit");
  TRYSYS(ioctl(*ofd,UI_SET_RELBIT,REL_Y),"set relbit");
  int i;
  for (i=0;i<KEY_MAX;i++) {
    if (test_bit(i,key_bitmask)||i<BTN_MISC) {
      TRYSYS(ioctl(*ofd,UI_SET_KEYBIT,i),"set keybit");
    }
  }
  TRYSYS(ioctl(*ofd,UI_SET_KEYBIT,BTN_MOUSE),"set keybit");

  struct uinput_user_dev ud = {0};
  strncpy(ud.name,"ukeyboard",UINPUT_MAX_NAME_SIZE);
  ud.id = id;
  TRYSYS(write(*ofd,&ud,sizeof(ud)),"write uinput");
  TRYSYS(ioctl(*ofd,UI_DEV_CREATE),"uinput create");
  TRYSYS(ioctl(*ifd,EVIOCGRAB),"ioctl grab");
}

void keyevent(int fd,__u16 key,__s32 value, struct input_event *eorig) {
  struct input_event e;
  e.time=eorig->time;
  e.type=EV_KEY;
  e.code=key;
  e.value=value;
  TRYSYS(write(fd,&e,sizeof(e)),"write");
}

void rely(int fd,__s32 value,struct input_event *eorig) {
  struct input_event e;
  e.time=eorig->time;
  e.type=EV_REL;
  e.code=REL_Y;
  e.value=value;
  TRYSYS(write(fd,&e,sizeof(e)),"write");
}

struct timeval add_nsec(const struct timeval *o, long offset) {
  long ous=o->tv_usec+offset/1000;
  struct timeval r;
  r.tv_sec=o->tv_sec+ous/1000000;
  r.tv_usec=ous%1000000;
  return r;
}

#define LONG_PRESS_DELAY (750*1000000L)
void translate(int fd,struct input_event *e) {
  static int home_pressed = 0;
  static int search_home_pressed = 0;
  static int search_pending = 0;
  if (e->code==KEY_HOME) {
    if (e->value) {
      home_pressed=1;
    } else {
      home_pressed=0;
      if (search_pending) {
	search_pending=0;
	struct timespec ts = {0,LONG_PRESS_DELAY};
	e->code=KEY_SEARCH;
	e->value=1;
	TRYSYS(write(fd,e,sizeof(*e)),"write");
	nanosleep(&ts,NULL);
	e->value=0;
	e->time = add_nsec(&e->time,LONG_PRESS_DELAY);
	TRYSYS(write(fd,e,sizeof(*e)),"write");
      } else {
	e->value=1;
	TRYSYS(write(fd,e,sizeof(*e)),"write");
	e->value=0;
	TRYSYS(write(fd,e,sizeof(*e)),"write");
      }
    }
  } else if (e->code==KEY_OPTION) { /* MENU key */
    if (home_pressed) {
      if (e->value) {
	search_home_pressed=1;
	search_pending=1;
      } else {
	search_home_pressed=0;
      }
    } else {
      if (!e->value && search_home_pressed) {
	search_home_pressed=0;
      } else {
	TRYSYS(write(fd,e,sizeof(*e)),"write");
      }
    }
  } else {
    TRYSYS(write(fd,e,sizeof(*e)),"write");
  }
}

void acquire_lock() {
  int fd;
  TRYSYS(fd = open("intrans-lock",O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP),"open lock");
  //  uid_t ruid = getuid();
  //  gid_t rgid = getgid();
  //  TRYSYS(fchown(fd,ruid,rgid),"chown lock");
  struct flock lock;
  lock.l_type=F_WRLCK;
  lock.l_whence=SEEK_SET;
  lock.l_start=0;
  lock.l_len=0;
  TRYSYS(fcntl(fd,F_SETLKW,&lock),"acquire write lock");
  char buf[1024];
  int n = snprintf(buf,1024,"%d\n",getpid());
  TRYSYS(write(fd,buf,n),"write pid");
  lock.l_type=F_RDLCK;
  TRYSYS(fcntl(fd,F_SETLKW,&lock),"degrade to read lock");
}

int main(int argc, char **argv) {
  int ifd, ofd; struct input_event e;
  init_devs(&ifd,&ofd);
  acquire_lock();
  puts("Intrans started");
  fflush(stdout);
  while(1) {
    if(read(ifd,&e,sizeof(e))==-1) { perror("read"); exit(1); }
    translate(ofd,&e);
  }
}
